package org.fictus

import scala.collection.mutable.ListBuffer
import org.fictus.internal.Method

class ExceptionWrapper(e: Exception) extends RuntimeException(e)

case class FictusResult[A](value: A)

trait Fictus extends MockProxy with Matchers with Logging {
    implicit val context = new MockContext(this)

    def mocks2[M1](implicit context: MockContext, m1: Manifest[M1]): (M1, M1) = {
        val nameBase = manifest.erasure.getSimpleName
        (mock(nameBase + "._1"), mock(nameBase + "._2"))
    }

    def mocks3[M1](implicit context: MockContext, m1: Manifest[M1]): (M1, M1, M1) = {
        val nameBase = manifest.erasure.getSimpleName
        (mock(nameBase + "._1"), mock(nameBase + "._2"), mock(nameBase + "._3"))
    }

    def mock[M1, M2](implicit context: MockContext, m1: Manifest[M1], m2: Manifest[M2]): (M1, M2) = (mock[M1], mock[M2])
    def mock[M1, M2, M3](implicit context: MockContext, m1: Manifest[M1], m2: Manifest[M2], m3: Manifest[M3]): (M1, M2, M3) = (mock[M1], mock[M2], mock[M3])
    def mock[M1, M2, M3, M4](implicit context: MockContext, m1: Manifest[M1], m2: Manifest[M2], m3: Manifest[M3], m4: Manifest[M4]): (M1, M2, M3, M4) = (mock[M1], mock[M2], mock[M3], mock[M4])

    /** @return a mock of the given type with a name of the given type. */
    def mock[A](implicit context: MockContext, manifest: Manifest[A]): A = mock(manifest.erasure.getSimpleName)

    /** @return a mock of the given type with the passed in name. */
    def mock[A](name: String)(implicit context: MockContext, manifest: Manifest[A]): A = {
        val mock = new Mock(name)
        createProxy(mock)
    }

    def missingBehaviors(expectation: Expectation[_]) = throw new PebkacError("\nExpected method is missing behaviors (returns or throws)\n" +
        expectation + " should look something like:\n" +
        "\t" + expectation + " --> <some return> OR\n" +
        "\t" + expectation + " --> <some exception>")

    def defaults(theDefaults: => Unit)(implicit context: MockContext) {
        def clearLastExpectation() = context.lastExpectation.isComplete match {
            case Some(expectation) => missingBehaviors(expectation)
            case None              => context.lastExpectation.clear
        }

        context.state.default
        clearLastExpectation()
        theDefaults
        clearLastExpectation()
        context.state.exitDefault
    }

    implicit def expect[A](expected: => A)(implicit context: MockContext): Expectation[A] = {
        context.lastExpectation.isComplete match {
            case Some(expectation) => missingBehaviors(expectation)
            case None =>
                expected
                context.lastExpectation.last[A]().get
        }
    }

    def test[A](unitFunction: => A)(implicit context: MockContext): A = {
        context.state.replay

        try {
            val result = unitFunction
            context.expectations.verifyAllExpectationsAreConsumed
            context.state.done
            result
        } catch {
            case e: ExceptionWrapper => throw e.getCause
        }
    }

    implicit def anyToAssertion[A](value: A): Assertion[A] = Assertion(value)

}

class Mock(name: String)(implicit context: MockContext) extends Invoke(name) {

    override val toString = name

    override def replay(method: Method, actualArgs: List[AnyRef]): AnyRef = {
        context.state.replay
        context.expectations.find(this, method) match {
            case Some(expectation) => replay(expectation, method, actualArgs)
            case None              => throw new PebkacError("Unexpected method called: " + name + "." + method.name + actualArgs.mkString("(", ", ", ")"))
        }
    }

    private def replay(expectation: RecordedBehavior, method: Method, actualArgs: List[AnyRef]) = {
        val matchers = expectation.expectation.matchers

        val argsMatch = matchers.zip(actualArgs).map {
            case (matcher, actual) => matcher.matches(actual)
        }

        context.logging.printSystemError(
            "REPLAY " + method.fullName +
                "\n  Matchers=" + matchers +
                "\n  Actual args=" + actualArgs +
                "\n  Behavior=" + expectation.behavior +
                "\n  Args Match=" + argsMatch)

        actualArgs.foreach { a =>
            context.logging.printSystemError("ACTUAL CLASS=" + a.getClass)
        }

        if (!argsMatch.isEmpty && argsMatch.contains(false)) {

            case class Match(matcher: Matcher, actualArg: AnyRef, matches: Boolean)

            val matches = argsMatch.zip(matchers).zip(actualArgs).map(tuple => Match(tuple._1._2, tuple._2, tuple._1._1))

            val expectedList = matches.map(m => m.matches match {
                case true  => m.matcher.expectedMsg
                case false => "->" + m.matcher.expectedMsg + "<-"
            })

            val actualList = matches.map(m => m.matches match {
                case true  => m.matcher.actualMsg(m.actualArg)
                case false => "->" + m.matcher.actualMsg(m.actualArg) + "<-"
            })

            val msg = "Expected arguments did not match the actual arguments!\n" +
                "expected: " + name + "." + method.name + expectedList.mkString("(", ", ", ")") + "\n" +
                "actual:   " + name + "." + method.name + actualList.mkString("(", ", ", ")")
            throw new ExpectationError(msg)
        }

        expectation.behavior.handle match {
            case ref: AnyRef => ref
            case value => value match {
                case i: Boolean => java.lang.Boolean.valueOf(i)
                case i: Byte    => java.lang.Byte.valueOf(i)
                case i: Char    => java.lang.Character.valueOf(i)
                case i: Short   => java.lang.Short.valueOf(i)
                case i: Int     => java.lang.Integer.valueOf(i)
                case i: Long    => java.lang.Long.valueOf(i)
                case i: Float   => java.lang.Float.valueOf(i)
                case i: Double  => java.lang.Double.valueOf(i)
            }
        }
    }

    override def record(method: Method, args: List[AnyRef]): AnyRef = {
        context.lastExpectation.store(Expectation(this, method, context.currentMatchers.get, args))
        Types.defaultValue(method.returnType)
    }
}

case class RecordedBehavior(expectation: Expectation[_], behavior: Behavior) {
    val methodName = expectation.methodName + behavior
}

object Expectation {
    def apply[A](mock: Mock, method: Method, matchers: List[Matcher], args: List[AnyRef])(implicit context: MockContext): Expectation[A] = {
        val parameterTypes = method.parameterTypes

        context.logging.printSystemError("MATCHERS=" + matchers)

        if (matchers.size > parameterTypes.size)
            throw new MatcherError("You have specified more matchers than arguments... only one matcher per argument will work.")
        if ((!args.isEmpty && !matchers.isEmpty) && (matchers.size != parameterTypes.size))
            throw new MatcherError("If you provide one matcher, you must provide matchers for all of the arguments! Sorry :(")

        val validMatchers = matchers match {
            case Nil => args.map(Equal(_))
            case _   => matchers
        }

        Expectation(mock, method, validMatchers)
    }
}
case class Expectation[A] private[fictus] (mock: Mock, method: Method, matchers: List[Matcher])(implicit context: MockContext) {

    def returns[B <: A](arg: B) = context.expectations.add(this, Returns(arg))
    def -->[B <: A](arg: B) = returns(arg)

    def throws(e: Exception) = context.expectations.add(this, Throws(e))
    def -->(e: Exception) = throws(e)

    def anyTimes() {
        context.expectations.convertLastBehaviorToAnyTimes()
    }

    val returnsUnit = method.returnType == classOf[Unit]

    val methodName = mock + "." + method.name + matchers.map(_.expectedMsg).mkString("(", ", ", ")")
    override val toString = methodName
}

private[fictus] class RecordedMatchers(context: MockContext) {
    private val matchers = new ListBuffer[Matcher]

    def append(matcher: Matcher) {
        matchers.append(matcher)
    }

    def get: List[Matcher] = {
        val theMatchers = matchers.toList
        matchers.clear
        theMatchers
    }
}

trait Expectations {
    def add[A](expectation: Expectation[A], behavior: Behavior): Expectation[A]
    def lastOption(): Option[RecordedBehavior]
    def convertLastBehaviorToAnyTimes(): Unit
    def verifyAllExpectationsAreConsumed: Unit
}

class Records(context: MockContext) extends Expectations {
    val defaults = new ExpectationsImpl("default", context.logging)
    val expectations = new ExpectationsImpl("expect", context.logging)

    def add[A](expectation: Expectation[A], behavior: Behavior): Expectation[A] = {
        context.state.current match {
            case DefaultRecording => defaults.add(expectation, behavior)
            case _                => expectations.add(expectation, behavior)
        }
    }

    def lastOption(): Option[RecordedBehavior] = context.state.current match {
        case DefaultRecording => defaults.lastOption
        case _                => expectations.lastOption
    }

    def convertLastBehaviorToAnyTimes() {
        context.state.current match {
            case DefaultRecording => defaults.convertLastBehaviorToAnyTimes
            case _                => expectations.convertLastBehaviorToAnyTimes
        }
    }

    def find(mock: Mock, method: Method): Option[RecordedBehavior] =
        expectations.findAndRemove(mock, method).orElse(defaults.find(mock, method))

    def verifyAllExpectationsAreConsumed {
        expectations.verifyAllExpectationsAreConsumed
    }

    override def toString = "Defaults=" + defaults + "\nExpectations=" + expectations
}

class ExpectationsImpl(name: String, logging: Logging) extends Expectations {
    val expectations = new ListBuffer[RecordedBehavior]

    def add[A](expectation: Expectation[A], behavior: Behavior): Expectation[A] = {
        expectation.returnsUnit match {
            case false =>
                logging.printVerbose("Recording " + "[" + name + "] :\n  " + expectation + "\n   " + behavior)
                expectations.append(RecordedBehavior(expectation, behavior))
            case true => behavior match {
                case throws @ Throws(_) =>
                    logging.printVerbose("   " + behavior)
                    expectations.remove(expectations.size - 1)
                    expectations.append(RecordedBehavior(expectation, behavior))
                case nop @ Nop() =>
                    logging.printVerbose("Recording " + "[" + name + "] :\n  " + expectation)
                    expectations.append(RecordedBehavior(expectation, behavior))
                case _           => throw new PebkacError("returns behavior is not allowed on mocked functions that return Unit.")
            }
        }

        expectation
    }

    def lastOption = expectations.lastOption

    def convertLastBehaviorToAnyTimes() = {
        expectations.lastOption match {
            case Some(last) =>
                expectations.remove(expectations.size - 1)
                expectations.append(last.copy(behavior = AnyTimes(last.behavior)))
                logging.printVerbose("    --> anyTimes()")
            case None => throw new PebkacError("FOREVER WHAT!?")
        }
    }

    def find(mock: Mock, method: Method): Option[RecordedBehavior] =
        expectations.find(expectation => (mock eq expectation.expectation.mock) && method == expectation.expectation.method).flatMap(recordedBehavior => {
            logging.printVerbose("Replaying " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
            Some(recordedBehavior)
        })

    def findAndRemove(mock: Mock, method: Method): Option[RecordedBehavior] = {
        expectations.zipWithIndex
            .find(expectation => (mock eq expectation._1.expectation.mock) && method == expectation._1.expectation.method)
            .map {
                case (recordedBehavior, index) => recordedBehavior.behavior match {
                    case AnyTimes(behavior) =>
                        logging.printVerbose("Replaying " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        recordedBehavior
                    case _ => {
                        val recordedBehavior = expectations.remove(index)
                        logging.printVerbose("Replaying " + "[" + name + "] :\n  " + recordedBehavior.expectation + "\n   " + recordedBehavior.behavior)
                        recordedBehavior
                    }
                }
            }
    }

    def verifyAllExpectationsAreConsumed {
        val regularExpections = expectations.toList.filterNot(_.behavior.isInstanceOf[AnyTimes])

        logging.printSystemError("LEFT OVER expectations=" + expectations)

        if (!regularExpections.isEmpty)
            throw new VerificationError("Expected methods were not invoked by the unit!\n" + regularExpections.map(_.methodName).mkString(" - ", "\n - ", ""))
    }

    override def toString = expectations.toString
}

class LastExpectation(expectations: Records, logging: Logging) {
    private var lastExpectation: Option[Expectation[_]] = None

    def store[A](expectation: Expectation[A]) = expectation.returnsUnit match {
        case false => lastExpectation = Some(expectation)
        case true =>
            lastExpectation = Some(expectation)
            expectations.add(expectation, Nop())
    }
    def last[A]() = lastExpectation.asInstanceOf[Option[Expectation[A]]]
    def clear() = lastExpectation = None
    def isComplete(): Option[Expectation[_]] = {

        logging.printSystemError(expectations.lastOption + " AND " + lastExpectation)
        logging.printSystemError(expectations.toString)

        (expectations.lastOption, lastExpectation) match {
            case (Some(recordedBehavior), Some(expectation)) if (expectation == recordedBehavior.expectation) => lastExpectation = None
            case _ =>
        }
        lastExpectation
    }
}

trait Logging {
    val printSystemErrors = false
    val printVerbose = false

    def printVerbose(s: String) {
        if (printVerbose)
            println(s)
    }

    def printSystemError(s: String) {
        if (printSystemErrors)
            println(s)
    }
}

trait LogSystemErrors { self: Logging =>
    override val printSystemErrors = true
}

trait LogVerbose { self: Logging =>
    override val printVerbose = true
}

trait State
case object Recording extends State
case object DefaultRecording extends State
case object Replaying extends State
case object Done extends State

class CurrentState {
    private var _state: State = Recording

    def current = _state

    def replay = current match {
        case Recording        => _state = Replaying
        case DefaultRecording => throw new StateError("Cannot call test in a defaults block")
        case Replaying        => // This is ok
        case Done             => throw new StateError("Cannot call anything after test is done :(")
    }

    def default = current match {
        case Recording        => _state = DefaultRecording
        case DefaultRecording => throw new StateError("Cannot have a defaults block inside a defaults block")
        case Replaying        => throw new StateError("Cannot have a defaults block in or after test")
        case Done             => throw new StateError("Cannot have a defaults block in or after test")
    }

    def exitDefault = current match {
        case Recording        => throw new WtfError("Recording -> exitDefault ??")
        case DefaultRecording => _state = Recording
        case Replaying        => throw new WtfError("Replaying -> exitDefault ??")
        case Done             => throw new WtfError("Done -> exitDefault ??")
    }

    def done = current match {
        case Recording        => throw new WtfError("Recording -> Done ??")
        case DefaultRecording => throw new WtfError("Default -> Done ??")
        case Replaying        => _state = Done
        case Done             => throw new WtfError("Done -> Done ??")
    }
}

class MockContext(val logging: Logging) extends Matchers {
    val state = new CurrentState
    val expectations = new Records(this)
    val currentMatchers = new RecordedMatchers(this)
    val lastExpectation = new LastExpectation(expectations, logging)
}